//---------------------------------------------------------------------------
#include <vcl\vcl.h>
//---------------------------------------------------------------------------
#pragma hdrstop
//---------------------------------------------------------------------------
#include "TMainForm.h"
#include "stdio.h"
#include "fcntl.h"
#include "io.h"
#include "dir.h"
#include "sys\stat.h"
//---------------------------------------------------------------------------
#if (__BORLANDC__ >= 0x0530) // BCB Version 3 or 4 only
    #pragma package(smart_init)
#endif
//---------------------------------------------------------------------------
#pragma resource "*.dfm"
//---------------------------------------------------------------------------
TMainForm *MainForm;
//---------------------------------------------------------------------------
#define word        unsigned short
#define dword       unsigned long
#define byte        unsigned char
//---------------------------------------------------------------------------
struct TPackItHeader
    {
    word        Version;
    dword       TotalFiles;
    TDateTime   DateCreated;
    };
//---------------------------------------------------------------------------
struct TPackItFileHeader
    {
    dword       FilePos;
    char        FileName[1024];
    dword       FileSize;
    };
//---------------------------------------------------------------------------
TStrings* ElementStrings( AnsiString SourceText, AnsiString Delimeter )
{
if (SourceText=="") return NULL;
TStrings* strings = new TStringList();
if (strings==NULL) return NULL;

/*
int pos;
int length = Delimeter.Length();
while (pos = SourceText.Pos(Delimeter))
    {
    strings->Add( SourceText.SubString(1,pos-1) );
    SourceText = SourceText.SubString(pos+length,SourceText.Length());
    }
return strings;
*/

char* source = SourceText.c_str();
char* delimeter = Delimeter.c_str();
int len = SourceText.Length();
int delimeter_length = Delimeter.Length();
int delimeter_count = 0;
AnsiString normal_text = "";
AnsiString possibly_normal_text = "";
for (int i=0;i<len;i++)
    {
    if (*source == *delimeter)  // if we found a delimeter
        {
        *delimeter++;           // get ready for the next one
        delimeter_count++;      // and remember how many we've completed
        if (delimeter_count==delimeter_length) // if we found a completed delimeter
            {
            // save the our built-up string
            strings->Add(normal_text);
            normal_text = "";
            possibly_normal_text = "";
            }
        else
            possibly_normal_text += *source;
        }
    else
        {
        normal_text += possibly_normal_text;
        possibly_normal_text = "";
        normal_text += *source;
        delimeter = Delimeter.c_str();
        delimeter_count = 0;
        }
    *source++;
    }
if (normal_text!="") strings->Add(normal_text);
return strings;
}
//---------------------------------------------------------------------------
AnsiString  ElementString( AnsiString SourceText, AnsiString Delimeter,
                            int ElementNumber )
{
TStrings* elems = ElementStrings( SourceText, Delimeter );
if (elems == NULL) return "";
AnsiString str = "";
if ((ElementNumber>=0) && (ElementNumber<=elems->Count-1))
    str = elems->Strings[ElementNumber];
delete elems;
return str;
}
//---------------------------------------------------------------------------
int ElementCount( AnsiString SourceText, AnsiString Delimeter )
{
TStrings* elems = ElementStrings( SourceText, Delimeter );
if (elems == NULL) return 0;
int count = elems->Count;
delete elems;
return count;
}
//---------------------------------------------------------------------------
__fastcall TMainForm::TMainForm(TComponent* Owner)
    : TForm(Owner)
{
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::ExitButtonClick(TObject *Sender)
{
if (SaveOnExit->Checked) SaveButtonClick(this);
Close();
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::SelectButtonClick(TObject *Sender)
{
for (int i=0;i<FileListBox1->Items->Count;i++) FileListBox1->Selected[i] = true;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::DeSelectButtonClick(TObject *Sender)
{
for (int i=0;i<FileListBox1->Items->Count;i++) FileListBox1->Selected[i] = false;
}
//---------------------------------------------------------------------------


void __fastcall TMainForm::BuildButtonClick(TObject *Sender)
{
if (Edit2->Text.Pos("*"))
    {
    ShowMessage( "You must select a valid PackIT file before building" );
    return;
    }

ContentsListBox->Items->Clear();
ContentsListBox->Update();

TPackItHeader header;
TPackItFileHeader fileheader;
AnsiString filename = PackItName->Caption;

chmod( filename.c_str(), S_IREAD | S_IWRITE );
if (FileExists(filename)) DeleteFile(filename);

int file = open(filename.c_str(),O_CREAT|O_RDWR|O_TRUNC|O_BINARY);
if (file == -1)
    {
    ShowMessage( "Cannot create new output file: '"+filename+"'" );
    SaveOnExit->Checked = false;
    return;
    }

// create temporary file header
header.Version = 0x0100;
header.TotalFiles = AddFilesMemo->Lines->Count;
header.DateCreated = TDateTime().CurrentDateTime();
write( file, &header, sizeof(TPackItHeader) );

// add all files
int num_added = 0;
for (int i=0;i<AddFilesMemo->Lines->Count;i++)
    {
    AnsiString tempfilename = AddFilesMemo->Lines->Strings[i];

    // try opening selected file
    int tempfile = open( tempfilename.c_str(), O_RDWR | O_BINARY );
    if (tempfile==-1)
        {
        ShowMessage("Could not open temporary file: '"+tempfilename+"'... skipped..." );
        continue;
        }

    // try allocating memory for the relevant file
    long templength = filelength( tempfile );
    char* tempbuffer = new char[templength];
    if (!tempbuffer)
        {
        close(tempfile);

        ShowMessage("Could not allocate temporary storage for file: '"+tempfilename+"'... skipped..." );
        SaveOnExit->Checked = false;
        continue;
        }

    // read entire file
    read( tempfile, tempbuffer, templength );
    close( tempfile );
    chmod( tempfilename.c_str(), S_IREAD | S_IWRITE );

    // add to the .pit file
    fileheader.FilePos = tell( file );
    fileheader.FileSize = templength;
    int len = tempfilename.Length();
    ZeroMemory( &fileheader.FileName, 255 );
    CopyMemory( &fileheader.FileName, tempfilename.SubString(1,len).c_str(), len );

    write( file, &fileheader, sizeof(TPackItFileHeader) );
    write( file, tempbuffer, templength );

    num_added++;

    ContentsListBox->Items->Add( tempfilename );
    ContentsListBox->ItemIndex = ContentsListBox->Items->Count-1;
    ContentsListBox->Update();

    delete[] tempbuffer;
    }

// make sure file contains correct # of entries
lseek( file, 0, SEEK_SET );

header.Version = 0x0100;
header.TotalFiles = num_added;
header.DateCreated = TDateTime().CurrentDateTime();

write( file, &header, sizeof(TPackItHeader) );
close(file);
chmod( filename.c_str(), S_IREAD | S_IWRITE );

if (num_added != AddFilesMemo->Lines->Count)
    {
    ShowMessage("PackIt file '"+filename+"' was NOT created correctly");
    SaveOnExit->Checked = false;
    }

// update the display appropriately
DirectoryListBox2->Update();
FileListBox2->Update();

for (int i=0;i<FileListBox2->Items->Count;i++)
    if (DirectoryListBox2->Directory+"\\"+FileListBox2->Items->Strings[i]==filename)
        {
        FileListBox2->ItemIndex = i;
        PackItName->Caption = filename;
        break;
        }

FileListBox2Change(this);
if (FileListBox2->ItemIndex!=-1) Edit2->Text = FileListBox2->Items->Strings[FileListBox2->ItemIndex];
}
//---------------------------------------------------------------------------


void __fastcall TMainForm::AddButtonClick(TObject *Sender)
{
for (int i=0;i<FileListBox1->Items->Count;i++)
    {
    if (FileListBox1->Selected[i])
        {
        bool found = false;
        for (int j=0;j<AddFilesMemo->Lines->Count;j++)
            if (AddFilesMemo->Lines->Strings[j]==DirectoryListBox1->Directory+"\\"+FileListBox1->Items->Strings[i] )
                {
                found = true;
                break;
                }
        if (!found) AddFilesMemo->Lines->Add( DirectoryListBox1->Directory+"\\"+FileListBox1->Items->Strings[i] );
        }
    }
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::DelButtonClick(TObject *Sender)
{
for (int i=0;i<FileListBox1->Items->Count;i++)
    {
    if (FileListBox1->Selected[i])
        {
        for (int j=0;j<AddFilesMemo->Lines->Count;j++)
            if (AddFilesMemo->Lines->Strings[j]==DirectoryListBox1->Directory+"\\"+FileListBox1->Items->Strings[i] )
                {
                AddFilesMemo->Lines->Delete(j);
                break;
                }
        }
    }
}
//---------------------------------------------------------------------------



void __fastcall TMainForm::Edit2KeyPress(TObject *Sender, char &Key)
{
Key_Was_Pressed = true;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::FormCreate(TObject *Sender)
{
Key_Was_Pressed = false;
Application->OnActivate = OnActivate;
Application->OnDeactivate = OnDeactivate;
}
//---------------------------------------------------------------------------


void __fastcall TMainForm::Edit2Exit(TObject *Sender)
{
if (Key_Was_Pressed)
    {
    Edit2->Text = Edit2->Text.UpperCase();
    if (Edit2->Text.Pos(".PIT")==0) Edit2->Text = Edit2->Text + ".PIT";
    }
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::CreateButtonClick(TObject *Sender)
{
if (Edit2->Text.Pos("*"))
    {
    ShowMessage( "You must select a valid PackIT file before creating." );
    return;
    }

if (Key_Was_Pressed)
    {
    AnsiString filename = PackItName->Caption;

    int file = open(filename.c_str(),O_CREAT|O_RDWR|O_TRUNC|O_BINARY);
    if ( file == -1 )
        {
        ShowMessage( "Cannot create output file: '"+filename+"'" );
        return;
        }

    TPackItHeader header;
    header.Version = 0x0100;
    header.TotalFiles = 0;
    header.DateCreated = TDateTime().CurrentDateTime();

    write( file, &header, sizeof(TPackItHeader) );

    close( file );

    FileListBox2->Update();
    FileListBox1->Update();
    ShowMessage("File Created");

    for (int i=0;i<FileListBox2->Items->Count;i++)
        if (DirectoryListBox2->Directory+"\\"+FileListBox2->Items->Strings[i]==filename)
            {
            FileListBox2->ItemIndex = i;
            break;
            }
    Edit2->Text = FileListBox2->Items->Strings[FileListBox2->ItemIndex];
    }

FileListBox2Change(this);
}
//---------------------------------------------------------------------------






void __fastcall TMainForm::ClearButtonClick(TObject *Sender)
{
AddFilesMemo->Lines->Clear();
}
//---------------------------------------------------------------------------


void __fastcall TMainForm::Edit2Change(TObject *Sender)
{
if (Edit2->Text!="*.PIT")
    {
    PackItName->Caption = DirectoryListBox2->Directory+"\\"+Edit2->Text;
    // try loading config file, if not possible, use existing contents
    AddFilesMemo->Lines->Clear();
    LoadConfigButtonClick(this);
    if (AddFilesMemo->Lines->Count==0) CopyButtonClick(this);
    }
else
    // resuming from (say) a alt-tab
    if (PackItName->Caption!="")
        Edit2->Text = ExtractFileName( PackItName->Caption );
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::FileListBox2Change(TObject *Sender)
{
ContentsListBox->Items->Clear();
if (FileListBox2->ItemIndex==-1) return;

AnsiString filename = PackItName->Caption;

chmod( filename.c_str(), S_IREAD | S_IWRITE );
int file = open(filename.c_str(),O_RDWR|O_BINARY);
if ( file == -1 )
    {
    ShowMessage( "Cannot open input file: '"+filename+"'" );
    return;
    }

TPackItHeader header;
TPackItFileHeader fileheader;

read( file, &header, sizeof(TPackItHeader) );

if (header.TotalFiles==0) ContentsListBox->Items->Add( "[EMPTY]" );
else
    for (dword i=0;i<header.TotalFiles;i++)
        {
        read( file, &fileheader, sizeof(TPackItFileHeader) );
        ContentsListBox->Items->Add( fileheader.FileName );

        // skip past file data to next header
        lseek( file, fileheader.FileSize, SEEK_CUR );
        }
close(file);
FormPaint(this);
}
//---------------------------------------------------------------------------


void __fastcall TMainForm::ExtractButtonClick(TObject *Sender)
{
if (Edit2->Text.Pos("*"))
    {
    ShowMessage( "You must select a valid PackIT file before extracting" );
    return;
    }

AnsiString filename = Edit2->Text;

// --------------------
// extract files...
// --------------------
TPackItHeader header;
TPackItFileHeader fileheader;

int file = open(filename.c_str(),O_RDWR|O_BINARY);
if (file == -1)
    {
    ShowMessage( "Could not open .PIT file." );
    SaveOnExit->Checked = false;
    return;
    }

// read PackIt file header (to get total # of files)
int error = read( file, &header, sizeof(TPackItHeader) );
if (error == -1)
    {
    close(file);
    ShowMessage( "Cannot read PackIt HEADER" );
    SaveOnExit->Checked = false;
    return;
    }

// change to relevant drive's root directory
AnsiString original_dir = GetCurrentDir();
AnsiString root_dir = "\\";
SetCurrentDir( root_dir );

// extract all PackIt files...
dword num_extracted = 0;
for (dword i=0;i<header.TotalFiles;i++)
    {
    // read file info
    read( file, &fileheader, sizeof(TPackItFileHeader) );

    // all files come in as :-
    //  eg.
    //      D:\TDx_Library\*.*
    // need to be
    //      O:\root_dir\TDx_Library\*.*

    AnsiString old_filename = fileheader.FileName;
    AnsiString old_directory = ElementString( old_filename, ":", 1);
    AnsiString new_filename = root_dir+old_directory;

    AnsiString just_drive = ElementString( new_filename, ":", 0);
    AnsiString just_filename = ExtractFileName( new_filename );
    AnsiString just_directories = ExtractFileDir( new_filename ).SubString( root_dir.Length()+1, new_filename.Length()-root_dir.Length());

//    ShowMessage("old_filename=="+old_filename+"\nold_directory=="+old_directory+"\n\nnew_filename=="+new_filename+
//    "\njust_drive=="+just_drive+"\njust_directories=="+just_directories+"\njust_filename=="+just_filename);

//    Percent->Caption = IntToStr(40*i/header.TotalFiles) + "% Complete";
    DifferentLabel->Caption = "Extracting "+just_filename;
    DifferentLabel->Update();
//    Application->ProcessMessages();

    // change to relevant drive's root directory
    SetCurrentDir( root_dir );

    // split up directories
    TStrings* dirs = ElementStrings( just_directories, "\\" );
    if (dirs==NULL)
        {
        close(file);
        ShowMessage( "The file '"+new_filename+"' could not be created because there's not enough memory available to create directories." );
        SaveOnExit->Checked = false;
        break;
        }

    // traverse the directory and create each subdir as required
    // ignore the first index, it's a null string
    for (int i=1;i<dirs->Count;i++)
        {
//            ShowMessage("Now Dir: "+GetCurrentDir()+"\ncreating sub: "+dirs->Strings[i]);
        // if the directory doesn't exist, create it
        if (!DirectoryExists( dirs->Strings[i] ))
            {
            CreateDir( dirs->Strings[i] );
            if (!DirectoryExists( dirs->Strings[i] ))
                {
                delete dirs;
                close(file);
                ShowMessage( "The file "+new_filename+" could not be created because one of it's directories ('"+dirs->Strings[i]+"') +could not be created." );
                break;
                }
            }
        // directory exists now

        // change to the newly created directory, ready to create the
        // next level down if required
        SetCurrentDir( dirs->Strings[i] );
        }

    // release directory strings
    delete dirs;

    // delete any existing file
    if (FileExists(just_filename))
        {
        // if for some reason, it is read-only.... change it to read-write
        chmod( just_filename.c_str(), S_IREAD | S_IWRITE );
        DeleteFile(just_filename);
        if (FileExists(just_filename))
            {
            close(file);
            ShowMessage( "There was an error when attempting to create the file '"+new_filename+"' because the original file could not be deleted.\n\nTry removing the file and re-installing.");
            // ignore this file, handle the next file
            SaveOnExit->Checked = false;
            break;
            }
        // continue below normally... // try creating new file
        }

    // try creating new file
    int tempfile = open( just_filename.c_str(), O_CREAT | O_TRUNC | O_WRONLY | O_BINARY );
    if (tempfile==-1)
        {
        close(file);
        close(tempfile);
        ShowMessage( "The file '"+new_filename+"' could not be created for some unknown reason, possibly lack of hard disk space, or an invalid filename?" );
        break;
        }

    // try allocating memory for the relevant file
    long templength = fileheader.FileSize;
    char* tempbuffer = new char[templength];
    if (!tempbuffer)
        {
        close(file);
        close(tempfile);
        delete[] tempbuffer;
        ShowMessage( "The file '"+new_filename+"' could not be created because there's not enough memory for it's temporary buffer." );
        SaveOnExit->Checked = false;
        break;
        }

    // read entire file
    read( file, tempbuffer, fileheader.FileSize );

    // save it to disk....
    write( tempfile, tempbuffer, templength );
    close( tempfile );

    num_extracted++;
    delete[] tempbuffer;

    // for some reason, it can be created as read-only....
    // make sure it is read-write
    chmod( new_filename.c_str(), S_IREAD | S_IWRITE );

    // restore to original drive/directory
    SetCurrentDir( original_dir );
    }
close(file);

// restore to original drive/directory
SetCurrentDir( original_dir );

if (num_extracted != header.TotalFiles)
    {
    ShowMessage( "ERROR: Not all the files were extracted correctly\n\n# of Files Files that were extracted = "+IntToStr(num_extracted)+"\n# of Files that should have been extracted = "+IntToStr((int)header.TotalFiles));
    SaveOnExit->Checked = false;
    }

// update the display appropriately
DirectoryListBox2->Update();
FileListBox2->Update();

for (int i=0;i<FileListBox2->Items->Count;i++)
    if (DirectoryListBox2->Directory+"\\"+FileListBox2->Items->Strings[i]==filename)
        {
        FileListBox2->ItemIndex = i;
        PackItName->Caption = filename;
        break;
        }

FileListBox2Change(this);
if (FileListBox2->ItemIndex!=-1) Edit2->Text = FileListBox2->Items->Strings[FileListBox2->ItemIndex];
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::SaveButtonClick(TObject *Sender)
{
if (Edit2->Text.Pos("*"))
    {
//    ShowMessage( "You must select a valid PackIT file before saving config" );
    return;
    }

AnsiString filename = PackItName->Caption.SubString( 1, PackItName->Caption.Pos(".")-1 ) + ".PFG";
try
    {
    AddFilesMemo->Lines->SaveToFile( filename );
    ShowMessage( "Config saved to file: '"+filename+"'" );
    }
catch (...)
    {
    ShowMessage( "Could not save configuration to file: '"+filename+"'" );
    }
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::LoadConfigButtonClick(TObject *Sender)
{
if (Edit2->Text.Pos("*"))
    {
    ShowMessage( "You must select a valid PackIT file before loading config" );
    return;
    }

AnsiString filename = PackItName->Caption.SubString( 1, PackItName->Caption.Pos(".")-1 ) + ".PFG";
AddFilesMemo->Lines->Clear();
if (FileExists(filename))
    try
        {
        AddFilesMemo->Lines->LoadFromFile( filename );
    //    ShowMessage( "Config loaded from file: '"+filename+"'" );
        }
    catch (...)
        {
        ShowMessage( "Could not load configuration from file: '"+filename+"'" );
        }
FormPaint(this);
}
//---------------------------------------------------------------------------


void __fastcall TMainForm::BitBtn1Click(TObject *Sender)
{
ContentsListBox->Items->Clear();    
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::CopyButtonClick(TObject *Sender)
{
for (int i=0;i<ContentsListBox->Items->Count;i++)
    {
    if (ContentsListBox->Items->Strings[i] == "[EMPTY]") continue;
    
    bool found = false;
    for (int j=0;j<AddFilesMemo->Lines->Count;j++)
        if (AddFilesMemo->Lines->Strings[j]==ContentsListBox->Items->Strings[i])
            {
            found = true;
            break;
            }
    if (!found) AddFilesMemo->Lines->Add( ContentsListBox->Items->Strings[i] );
    }
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::Label6Click(TObject *Sender)
{
OverwriteCheckBox->Checked = !OverwriteCheckBox->Checked;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::FormActivate(TObject *Sender)
{
DirectoryListBox1->Update();
DirectoryListBox2->Update();
FileListBox1->Update();
FileListBox2->Update();

// re-select the appropriate item
if (LastPackItName!="")
    {
    bool found = false;
    for (int i=0;i<FileListBox2->Items->Count;i++)
        if (DirectoryListBox2->Directory + "\\" + FileListBox2->Items->Strings[i] == LastPackItName)
            {
            FileListBox2->ItemIndex = i;
            found = true;
            break;
            }
    PackItName->Caption = LastPackItName;
    LastPackItName = "";
    if (found) FileListBox2Change(this);
    }
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::FormDeactivate(TObject *Sender)
{
if (LastPackItName=="") LastPackItName = PackItName->Caption;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::FormPaint(TObject *Sender)
{
if (AddFilesMemo->Lines->Text!=ContentsListBox->Items->Text)
    {
    DifferentLabel->Caption = "WARNING: Content of PIT differs from CFG";
    }
else
    DifferentLabel->Caption = "";
PITLabel->Caption = IntToStr(ContentsListBox->Items->Count);
PFGLabel->Caption = IntToStr(AddFilesMemo->Lines->Count);
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::AddFilesMemoExit(TObject *Sender)
{
FormPaint(this);    
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::ComboBox1Change(TObject *Sender)
{
if (AddFilesMemo->Lines->Count>0)
    {
    AnsiString drive = ComboBox1->Items->Strings[ComboBox1->ItemIndex];
    for (int i=0;i<AddFilesMemo->Lines->Count;i++)
        AddFilesMemo->Lines->Strings[i] = drive + AddFilesMemo->Lines->Strings[i].SubString(3,AddFilesMemo->Lines->Strings[i].Length()-2);
    }
}
//---------------------------------------------------------------------------

